import os

import chess.engine

# make sure stockfish environment variable exists
if "STOCKFISH_EXECUTABLE" not in os.environ:
    raise KeyError(
        'This bot requires an environment variable called "STOCKFISH_EXECUTABLE" '
        "pointing to the Stockfish executable"
    )

# make sure there is actually a file
STOCKFISH_EXECUTABLE = os.getenv("STOCKFISH_EXECUTABLE")
if not os.path.exists(STOCKFISH_EXECUTABLE):
    raise ValueError(f'No stockfish executable found at "{STOCKFISH_EXECUTABLE}"')


def create_engine():
    engine = chess.engine.SimpleEngine.popen_uci(STOCKFISH_EXECUTABLE, setpgrp=True)
    return engine



"""Asynchronous pygame UI for displaying each players' MHT view of the game during a replay

Work in progress!

For simplicity, all asynchronous functions are infinite while loops that yield control when
possible to allow other tasks to run. Information is passed between tasks using queues and by
updating shared state. The game loop is based on the pygame event loop so the entire process can
be terminated when that loop returns (e.g. when the user closes the window).

The desired behavior is for the window to appear immediately and for navigation and rendering to
appear even before the MHT calculations are complete. To accomplish this, we create a surface to
store each board view and the status computing and rendering the boards. We also store a message
to be displayed per turn. Then the animation task has only to render the current surfaces at each
step (and we could possibly set a timestamp for the last change to avoid blit-ing unchanged
surfaces). A separate task runs to compute the MHT views, blit the pieces to each surface,
and update the status information.
"""

import asyncio
import contextlib
import time
from typing import List, Optional, Tuple

from reconchess import GameHistory

with contextlib.redirect_stdout(None):
    import pygame

import chess

from reconchess_tools.ui import PIECE_IMAGES, draw_boards, draw_empty_board
from reconchess_tools.utilities import (
    possible_requested_moves,
    simulate_move,
    simulate_sense,
)

SENSE, MOVE = False, True


class Replay:
    def __init__(self, history_string: str):

        pygame.init()
        pygame.display.set_caption("Reconchess MHT Replay")
        pygame.display.set_icon(
            pygame.transform.scale(
                PIECE_IMAGES[chess.Piece(chess.KING, chess.WHITE)], (32, 32)
            )
        )

        self.header_font = pygame.font.SysFont(pygame.font.get_default_font(), 28)
        self.body_font = pygame.font.SysFont(pygame.font.get_default_font(), 20)
        self.body_spacing = 20
        self.board_font = pygame.font.SysFont(pygame.font.get_default_font(), 20)

        self.background_color = (45, 48, 50)
        self.header_color = (250, 250, 250)
        self.body_color = (160, 160, 160)

        self.square_size = 40
        self.margin = 10
        self.board_size = self.square_size * 8
        self.width = self.square_size * 16 + self.margin * 3
        self.height = self.width
        self.screen = pygame.display.set_mode((self.width, self.height))
        self.screen.fill(self.background_color)

        self.action_index = 0

        self.history = tuple(history_string.strip().lower().split())
        self.history_string = " ".join(
            self.history
        )  # to clean up possible multiple spaces
        self.num_actions = len(self.history)
        self.num_moves = self.num_actions // 2
        self.num_moves_by_white = (self.num_moves + 1) // 2
        self.num_moves_by_black = self.num_moves // 2

        board = chess.Board()
        self.views: List[View] = [View(board, self.board_font, self.board_size)]

        # Compute the true board states synchronously since that is fast
        history_iter = iter(self.history)
        try:
            while True:
                # Sense step
                next(history_iter)
                # Move step
                requested_move = chess.Move.from_uci(next(history_iter))
                taken_move, capture_square = simulate_move(board, requested_move)
                board.push(taken_move)
                self.views.append(View(board, self.board_font, self.board_size))

        except StopIteration:
            pass

        self.winner = not board.turn
        self.win_reason = "timeout" if board.king(board.turn) else "king capture"

    @classmethod
    def from_history(cls, history: GameHistory) -> "Replay":
        actions = []
        for turn in history.turns():
            sense = history.sense(turn)
            actions.append("00" if sense is None else chess.SQUARE_NAMES[sense])
            actions.append((history.requested_move(turn) or chess.Move.null()).uci())
        actions = " ".join(actions)
        return Replay(actions)

    async def play(self):
        task_mht = asyncio.create_task(self.update_mht())
        task_update = asyncio.create_task(self.update_view())
        await self.respond_to_events()
        task_mht.cancel()
        task_update.cancel()
        pygame.quit()

    def play_sync(self):
        asyncio.run(self.play())

    async def update_mht(self):
        history_iter = iter(self.history)
        board = chess.Board()
        active, waiting = AsyncMultiHypothesisTracker(), AsyncMultiHypothesisTracker()
        turn_index = 0
        num_boards = [1, 1]

        requested_move = (
            taken_move
        ) = capture_square = piece_moved = piece_captured = None
        x = y = 10
        view = self.views[0]

        surface_sense = pygame.Surface([self.square_size * 3] * 2, pygame.SRCALPHA)
        surface_sense.fill((205, 205, 255, 85))

        surface_capture = pygame.Surface([self.square_size] * 2, pygame.SRCALPHA)
        surface_capture.fill((255, 0, 0, 50))

        # TODO: Split white and black MHT views in case one explodes
        try:
            while True:
                view = self.views[turn_index]
                if board.turn == chess.WHITE:
                    view.surface_white = draw_boards(
                        active.boards, self.board_size, self.board_font
                    )
                    view.surface_black = draw_boards(
                        waiting.boards, self.board_size, self.board_font
                    )
                else:
                    view.surface_white = draw_boards(
                        waiting.boards, self.board_size, self.board_font
                    )
                    view.surface_black = draw_boards(
                        active.boards, self.board_size, self.board_font
                    )
                # Shade capture square
                if capture_square is not None:
                    x = self.square_size * chess.square_file(capture_square)
                    y = self.board_size - self.square_size * (
                        chess.square_rank(capture_square) + 1
                    )
                    view.surface_true.blit(surface_capture, (x, y))
                    view.surface_white.blit(surface_capture, (x, y))
                    view.surface_black.blit(surface_capture, (x, y))
                view.updated_at = time.monotonic()
                await asyncio.sleep(0)

                # Update info
                view.surface_info.fill(self.background_color)
                if turn_index == 0:
                    info = ["White to sense on turn 1"]
                else:
                    info = [
                        f"{chess.COLOR_NAMES[not board.turn].capitalize()} "
                        f"requested to move {chess.PIECE_NAMES[piece_moved.piece_type]} "
                        f"{requested_move} on turn "
                        f"{(turn_index - 1) // 2 + 1}, which",
                        f"    resulted in move {taken_move} and "
                        + (
                            f"the capture of the {chess.PIECE_NAMES[piece_captured.piece_type]} at "
                            f"{chess.SQUARE_NAMES[capture_square]}"
                            if piece_captured
                            else "no capture"
                        ),
                        f"# possible boards for {chess.COLOR_NAMES[not board.turn]}: "
                        f"{num_boards[not board.turn]:,.0f} -> {len(waiting.boards):,.0f} "
                        f"(Δ = {len(waiting.boards) - num_boards[not board.turn]:+,.0f})",
                        f"# possible boards for {chess.COLOR_NAMES[board.turn]}: "
                        f"{num_boards[board.turn]:,.0f} -> {len(active.boards):,.0f} "
                        f"(Δ = {len(active.boards) - num_boards[board.turn]:+,.0f})",
                        "",
                        f"{chess.COLOR_NAMES[board.turn].capitalize()} to sense on turn {turn_index // 2 + 1}",
                    ]
                x = y = 10
                for line in info:
                    view.surface_info.blit(
                        self.body_font.render(line, True, self.body_color), (x, y)
                    )
                    y += 20
                num_boards[board.turn] = len(active.boards)
                num_boards[not board.turn] = len(waiting.boards)
                await asyncio.sleep(0)

                # Sense step
                square = next(history_iter)
                square = None if square == "00" else chess.parse_square(square)
                result = simulate_sense(board, square)
                await active.sense(square, result)
                view.surface_after_sense = draw_boards(
                    active.boards, self.board_size, self.board_font
                )
                # Shade sensed squares
                if square is not None:
                    x = self.square_size * (chess.square_file(square) - 1)
                    y = self.board_size - self.square_size * (
                        chess.square_rank(square) + 2
                    )
                    view.surface_after_sense.blit(surface_sense, (x, y))
                view.updated_at = time.monotonic()
                await asyncio.sleep(0)

                # Update info
                view.surface_info_after_sense.fill(self.background_color)
                info = [
                    f"{chess.COLOR_NAMES[board.turn].capitalize()} "
                    + (
                        f"sensed at {chess.SQUARE_NAMES[square]}"
                        if square is not None
                        else "did not sense"
                    )
                    + f" on turn {turn_index // 2 + 1}",
                    f"# possible boards for {chess.COLOR_NAMES[board.turn]}: "
                    f"{num_boards[board.turn]:,.0f} -> {len(active.boards):,.0f} "
                    f"(Δ = {len(active.boards) - num_boards[board.turn]:+,.0f})",
                    "",
                    f"{chess.COLOR_NAMES[board.turn].capitalize()} to move on turn {(turn_index + 1) // 2 + 1}",
                ]
                x = y = 10
                for line in info:
                    view.surface_info_after_sense.blit(
                        self.body_font.render(line, True, self.body_color), (x, y)
                    )
                    y += 20
                num_boards[board.turn] = len(active.boards)
                await asyncio.sleep(0)

                # Move step
                requested_move = chess.Move.from_uci(next(history_iter))
                taken_move, capture_square = simulate_move(board, requested_move)
                piece_moved = (
                    board.piece_at(requested_move.from_square)
                    if requested_move
                    else None
                )
                piece_captured = (
                    None if capture_square is None else board.piece_at(capture_square)
                )
                await active.move(requested_move, taken_move, capture_square)
                await waiting.op_move(capture_square)
                board.push(taken_move)
                turn_index += 1
                active, waiting = waiting, active

        except StopIteration:
            pass

        info = [
            "",
            f"{chess.COLOR_NAMES[self.winner].capitalize()} wins by {self.win_reason}!",
        ]
        for line in info:
            (
                view.surface_info_after_sense
                if self.num_actions % 2
                else view.surface_info
            ).blit(self.body_font.render(line, True, self.body_color), (x, y))
            y += 20

    async def respond_to_events(self):
        while True:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    return
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_LEFT:
                        self.action_index = max(0, self.action_index - 1)
                    elif event.key == pygame.K_RIGHT:
                        self.action_index = min(self.num_actions, self.action_index + 1)
                    elif event.key == pygame.K_ESCAPE:
                        pygame.quit()
                        return
            await asyncio.sleep(0.005)

    async def update_view(self):
        dt = 1 / 60
        actual_fps = 1 / dt
        alpha = 0.98
        last_update_time = time.time()
        while True:
            view = self.views[self.action_index // 2]
            surface_true = view.surface_true
            surface_white = view.surface_white
            surface_black = view.surface_black
            surface_info = view.surface_info
            if self.action_index % 2:  # has sensed
                surface_info = view.surface_info_after_sense
                if view.active_player == chess.WHITE:
                    surface_white = view.surface_after_sense
                else:
                    surface_black = view.surface_after_sense

            self.screen.blit(surface_true, (self.margin, self.margin))
            self.screen.blit(
                surface_white,
                (self.margin, self.margin * 2 + self.square_size * 8),
            )
            self.screen.blit(
                surface_black,
                (
                    self.margin * 2 + self.square_size * 8,
                    self.margin * 2 + self.square_size * 8,
                ),
            )
            self.screen.blit(
                surface_info, (self.margin * 2 + self.square_size * 8, self.margin)
            )
            pygame.display.flip()
            current_time = time.time()
            actual_fps = alpha * actual_fps + (1 - alpha) / max(
                1e-3, current_time - last_update_time
            )
            pygame.display.set_caption(f"Reconchess MHT Replay ({actual_fps:.1f} fps)")
            await asyncio.sleep(last_update_time + dt - current_time)
            # await asyncio.sleep(dt)
            last_update_time = current_time


class View:
    def __init__(self, true_board, font, width):
        self.surface_true = draw_boards([true_board], width, font)
        self.surface_white = draw_empty_board(font, width)
        self.surface_black = draw_empty_board(font, width)
        self.surface_after_sense = draw_empty_board(font, width)
        self.surface_info = pygame.Surface((width, width))
        self.surface_info_after_sense = pygame.Surface((width, width))
        self.active_player = true_board.turn
        self.updated_at = time.monotonic()


def board_fingerprint(board: chess.Board):
    return (
        board.turn,
        *board.occupied_co,
        board.kings,
        board.queens,
        board.bishops,
        board.knights,
        board.rooks,
        board.pawns,
        board.castling_rights,
        board.ep_square,
    )


class AsyncMultiHypothesisTracker:
    def __init__(self):
        self.boards = [chess.Board()]

    async def sense(self, square: chess.Square, result: List[Tuple[int, chess.Piece]]):
        new_boards = []
        for board in self.boards:
            if simulate_sense(board, square) == result:
                new_boards.append(board)
            await asyncio.sleep(0)
        self.boards = new_boards

    async def move(
        self,
        requested_move: chess.Move,
        taken_move: chess.Move,
        capture_square: Optional[chess.Square],
    ):
        new_boards = []
        for board in self.boards:
            if simulate_move(board, requested_move) == (taken_move, capture_square):
                board.push(taken_move)
                new_boards.append(board)
            await asyncio.sleep(0)
        self.boards = new_boards

    async def op_move(self, capture_square: Optional[chess.Square]):
        new_boards = {}
        for board in self.boards:
            for requested_move in possible_requested_moves(board):
                taken_move, simulated_capture_square = simulate_move(
                    board, requested_move
                )
                if simulated_capture_square == capture_square:
                    new_board = board.copy(stack=False)
                    new_board.push(taken_move)
                    new_boards[board_fingerprint(new_board)] = new_board
                await asyncio.sleep(0)
        self.boards = list(new_boards.values())


def _main():
    Replay(
        """
        E6 e2e4 E3 c7c6 G7 d2d4 D2 d8b6 C5 b1c3 B2 b6b4 B4 g1f3 B2 d7d5 C7 a2a3 B4 b4a5 E7 a3b4 C4 b8d7 F3 a3b4 E2 a5a6 B3 c3b5 C2 a6a5 B7 b5c7 G2 a5e1
        """
    ).play_sync()


if __name__ == "__main__":
    _main()



import random
from math import sqrt
from typing import List

import chess
import pkg_resources
import pygame

LIGHT_COLOR = (240, 217, 181)
DARK_COLOR = (181, 136, 99)

PIECE_IMAGES = {}
for color in chess.COLORS:
    for piece_type in chess.PIECE_TYPES:
        piece = chess.Piece(piece_type, color)

        img_path = "res/{}/{}.png".format(chess.COLOR_NAMES[color], piece.symbol())
        full_path = pkg_resources.resource_filename("reconchess", img_path)

        img = pygame.image.load(full_path)
        PIECE_IMAGES[piece] = img


def draw_empty_board(font: pygame.font.SysFont, w) -> pygame.Surface:
    surface = pygame.Surface((w, w))
    pygame.draw.rect(surface, LIGHT_COLOR, (0, 0, w, w))
    sw = w / 8

    for dark_square in chess.SquareSet(chess.BB_DARK_SQUARES):
        sx = sw * chess.square_file(dark_square)
        sy = w - sw - sw * chess.square_rank(dark_square)
        pygame.draw.rect(surface, DARK_COLOR, (sx, sy, sw, sw))

    example_label = font.render("a", True, (0, 0, 0))
    rect = example_label.get_rect()
    for i in range(0, 8, 2):
        surface.blit(
            font.render(chess.FILE_NAMES[i], True, LIGHT_COLOR),
            (sw * i, w - rect.height),
        )
        surface.blit(
            font.render(chess.RANK_NAMES[i], True, DARK_COLOR),
            (w - rect.width, w - sw - sw * i),
        )
    for i in range(1, 8, 2):
        surface.blit(
            font.render(chess.FILE_NAMES[i], True, DARK_COLOR),
            (sw * i, w - rect.height),
        )
        surface.blit(
            font.render(chess.RANK_NAMES[i], True, LIGHT_COLOR),
            (w - rect.width, w - sw - sw * i),
        )

    return surface


def draw_pieces(board: chess.Board, w, alpha) -> pygame.Surface:
    surface = pygame.Surface((w, w), pygame.SRCALPHA)
    sw = w / 8
    for square in chess.SQUARES:
        piece = board.piece_at(square)
        if piece is not None:
            image = pygame.transform.scale(PIECE_IMAGES[piece], (int(sw), int(sw)))
            s = pygame.Surface((sw, sw), pygame.SRCALPHA)
            s.fill((255, 255, 255, alpha))
            s.blit(image, (0, 0), special_flags=pygame.BLEND_RGBA_MULT)
            x = sw * chess.square_file(square)
            y = w - sw - sw * chess.square_rank(square)
            surface.blit(s, (x, y))
    return surface


def draw_boards(
    boards: List[chess.Board], w, font: pygame.font.SysFont, max_boards=10_000
) -> pygame.Surface:
    if len(boards) > max_boards:
        boards = random.sample(boards, max_boards)
    surface = draw_empty_board(font, w)
    alpha = max(1, int(255 / sqrt(len(boards))))
    for board in boards:
        surface.blit(draw_pieces(board, w, alpha), (0, 0))
    return surface



from typing import Iterable, List, Optional, Tuple

import chess
from reconchess.utilities import (
    capture_square_of_move,
    is_illegal_castle,
    move_actions,
    revise_move,
    without_opponent_pieces,
)

_BACKRANK_SQUARES = chess.SquareSet(chess.BB_BACKRANKS)


def simulate_sense(
    board: chess.Board, square: Optional[chess.Square]
) -> List[Tuple[chess.Square, Optional[chess.Piece]]]:
    if square is None:
        return []
    assert square in list(chess.SQUARES), f"{square} is not a valid square."
    rank, file = chess.square_rank(square), chess.square_file(square)
    sense_result = []
    for delta_rank in [-1, 0, 1]:
        for delta_file in [-1, 0, 1]:
            if 0 <= rank + delta_rank <= 7 and 0 <= file + delta_file <= 7:
                sense_square = chess.square(file + delta_file, rank + delta_rank)
                sense_result.append((sense_square, board.piece_at(sense_square)))
    return sense_result


def simulate_move(board, move: chess.Move) -> Tuple[chess.Move, Optional[int]]:
    if move:
        taken_move = revise_move(board, move) or chess.Move.null()
        return taken_move, capture_square_of_move(board, taken_move)
    return move, None


def possible_requested_moves(board: chess.Board) -> Iterable[chess.Move]:
    yield from move_actions(board)
    yield chess.Move.null()


def possible_taken_moves(board: chess.Board) -> Iterable[chess.Move]:
    for move in board.pseudo_legal_moves:
        yield move
    for move in without_opponent_pieces(board).generate_castling_moves():
        if not is_illegal_castle(board, move):
            yield move
    yield chess.Move.null()



import chess
import click
import reconchess

from reconchess_tools.ui.replay import Replay


@click.group()
def cli():
    pass


@cli.command()
@click.argument("white_path", type=str)
@click.argument("black_path", type=str)
def bot_match(white_path, black_path):
    game = reconchess.LocalGame(900)

    _, white = reconchess.load_player(white_path)
    _, black = reconchess.load_player(black_path)

    winner_color, win_reason, history = reconchess.play_local_game(
        white(), black(), game=game
    )
    winner = "Draw" if winner_color is None else chess.COLOR_NAMES[winner_color]

    print("Game Over!")
    print(f"Winner: {winner}! ({win_reason})")

    Replay.from_history(history).play_sync()


if __name__ == "__main__":
    cli()



from collections import defaultdict
from typing import List, Optional, Tuple

import chess

from reconchess_tools.strategy import SENSE_SQUARES
from reconchess_tools.utilities import (
    possible_requested_moves,
    simulate_move,
    simulate_sense,
)


def board_fingerprint(board: chess.Board):
    """Compute a fingerprint for fast board comparisons

    This fingerprint is a tuple of integers and booleans that contains the same information as the
    extended position description (EPD). It does not contain all of the information in the FEN, e.g.
    half-move counter, because those are not significant in reconchess. Two boards have the same
    fingerprint, they are identical as far as reconchess is concerned, including allowing the same
    requested moves, and having the same results for any sense or move action.
    """
    return (
        board.turn,
        *board.occupied_co,
        board.kings,
        board.queens,
        board.bishops,
        board.knights,
        board.rooks,
        board.pawns,
        board.castling_rights,
        board.ep_square,
    )


class MultiHypothesisTracker:
    """An object to keep track of the possible true board states in a reconchess game

    The MultiHypothesisTracker (MHT) maintains a list of the possible true board positions. At the
    start of the game that list contains only the starting position. Whenever a player receives a
    sense result, that list may be filtered to include only the boards for which the simulated sense
    result matches. Likewise, whenever a player receives a move result, the list may be filtered to
    include only those boards for which the requested move would have resulted in the same taken
    move and capture square, after which the taken move must be pushed to all boards to keep them
    up to date. And lastly, after an opponent moves, each board in the list must be expanded into
    the result of all possible taken moves for which the capture square was the same as observed.
    Because that may result in duplicate positions, the list should be filtered to ensure
    uniqueness.

    Because the board list and the boards themselves are mutable, you must take care not to
    unintentionally alter them. For example, you may need copy a board from this list if you want
    to alter it (or you can ensure you undo any moves made on it). It is possible that the list will
    grow too large to be reasonably sustained. To prevent that, you may slice the list and all the
    methods below will still work. If you do so, you may discard the true board state, in which
    case it is possible to reveal information that contradicts all remaining boards and leaves the
    list of possible boards empty. Therefore, if you discard boards to limit memory requirements, be
    sure to handle the edge case of an empty board set.

    It is sometimes beneficial to speculate outcomes before making a decision. For example, to
    imagine all possible outcomes of a sense decision across all possible (or reasonable) choices.
    The speculate_sense method does just that and stores the result in the sense_speculation
    property. If present, this is used in the sense step rather than recomputing simulated sense
    results. The sense_speculation property can be input to various functions in the strategy module
    to aid sense decision making. It is reset to None after it is used in the sense step.
    """

    def __init__(self):
        self.boards = [chess.Board()]

        # An optional nested map of subsequent boards given a sense square and sense result
        self.sense_speculation = None

        # TODO speculation
        #  - For each of my move and opponent move, add a method to calculate all possible outcomes
        #    without the prior information.
        #  - Have existing methods do a lookup on speculation results if present, then delete them.

    def reset(self):
        self.boards = [chess.Board()]

    def speculate_sense(self, sense_squares=SENSE_SQUARES):
        self.sense_speculation = {}
        for square in sense_squares:
            self.sense_speculation[square] = sense_results = defaultdict(list)
            for board in self.boards:
                sense_results[tuple(simulate_sense(board, square))].append(board)

    def sense(self, square: chess.Square, sorted_result: List[Tuple[int, chess.Piece]]):
        if self.sense_speculation is not None:
            self.boards = self.sense_speculation[square][tuple(sorted_result)]
            self.sense_speculation = None
        else:
            self.boards = [
                board
                for board in self.boards
                if simulate_sense(board, square) == sorted_result
            ]

    def move(
        self,
        requested_move: chess.Move,
        taken_move: chess.Move,
        capture_square: Optional[chess.Square],
    ):
        self.boards = [
            board
            for board in self.boards
            if simulate_move(board, requested_move) == (taken_move, capture_square)
        ]
        for board in self.boards:
            board.push(taken_move)

    def op_move(self, capture_square: Optional[chess.Square]):
        new_boards = {}
        for board in self.boards:
            for requested_move in possible_requested_moves(board):
                taken_move, simulated_capture_square = simulate_move(
                    board, requested_move
                )
                if simulated_capture_square == capture_square:
                    new_board = board.copy(stack=False)
                    new_board.push(taken_move)
                    new_boards[board_fingerprint(new_board)] = new_board
        self.boards = list(new_boards.values())


if __name__ == "__main__":
    board = chess.Board()
    mht = MultiHypothesisTracker()

    board.push(chess.Move.from_uci("g1h3"))
    mht.op_move(None)
    mht.sense(chess.C2, simulate_sense(board, chess.C2))
    mht.move(chess.Move.from_uci("d7d5"), chess.Move.from_uci("d7d5"), None)
    board.push(chess.Move.from_uci("d7d5"))

    board.push(chess.Move.from_uci("h3f4"))
    mht.op_move(None)
    mht.sense(chess.F2, simulate_sense(board, chess.F2))
    mht.move(chess.Move.from_uci("e7e5"), chess.Move.from_uci("e7e5"), None)
    board.push(chess.Move.from_uci("e7e5"))

    board.push(chess.Move.from_uci("f4h5"))
    mht.op_move(None)
    mht.sense(chess.C2, simulate_sense(board, chess.C2))
    mht.move(chess.Move.from_uci("g8f6"), chess.Move.from_uci("g8f6"), None)
    board.push(chess.Move.from_uci("g8f6"))

    board.push(chess.Move.from_uci("h5f6"))
    mht.op_move(chess.F6)



import io
import os

from setuptools import find_packages, setup

# Package meta-data.
NAME = "reconchess-tools"
DESCRIPTION = "Tools for developing bots to play reconchess."
URL = "https://github.com/raaperrotta/reconchess-tools"
EMAIL = "raaperrotta@gmail.com"
AUTHOR = "Robert Perrotta"
REQUIRES_PYTHON = ">=3.6.0"
VERSION = "0.2.1"

# What packages are required for this module to be executed?
REQUIRED = [
    "reconchess",
    "tqdm",
    "click",
]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# If you change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = "\n" + f.read()

# Load the package's __version__.py module as a dictionary.
about = {"__version__": VERSION}

# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests"]),
    entry_points={
        "console_scripts": ["reconchess-tools=reconchess_tools.cli.main:cli"],
    },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



import random
from collections import defaultdict
from typing import List, Optional, Tuple

import chess.engine
from reconchess import Color, GameHistory, Player, WinReason
from tqdm import tqdm

from reconchess_tools.mht import MultiHypothesisTracker
from reconchess_tools.stockfish import create_engine
from reconchess_tools.strategy import (
    certain_win,
    minimax_sense,
    non_dominated_sense_by_own_pieces,
)
from reconchess_tools.utilities import simulate_move


class MhtBot(Player):
    def __init__(self):
        # We use the MHT object to handle all uncertainty tracking. In the methods below, we call
        # the appropriate update methods of the MHT object after which its boards property contains
        # the list of all chess boards that might be the true state of the game board. It is
        # important to be aware that both that list and the boards within it are mutable. It is the
        # responsibility of the user to avoid mutating those except intentionally, for example, we
        # slice the list below to prevent if from growing too large, even though that might mean we
        # lose track of the true board state.
        self.mht = MultiHypothesisTracker()
        # We use Stockfish (though this could be any UCI-compliant engine) to analyze the possible
        # boards. After handling boards that are not valid in regular chess (i.e. the opponent king
        # can be captured, or we are in checkmate) we ask stockfish to suggest a few moves, which we
        # aggregate using a variation of ranked-choice-voting. (More on that below.)
        self.engine = create_engine()

        self.color = None
        self.turn_num = None

    def handle_game_start(
        self, color: chess.Color, board: chess.Board, opponent_name: str
    ):
        self.color = color
        # Initializing this to -1 makes handling the first turn easier
        self.turn_num = -1
        self.mht.reset()

    def handle_opponent_move_result(
        self, captured_my_piece: bool, capture_square: Optional[chess.Square]
    ):
        self.turn_num += 1
        # This is called even before the first action. In that case, do nothing.
        if self.turn_num == 0 and self.color == chess.WHITE:
            return
        # Though we risk losing track of the true board, to avoid the number of possible boards
        # growing too large we limit its size here. Expanding the possible boards to account for all
        # possible opponent moves is the most demanding step in the MHT processing so it is critical
        # that we ensure this next step is tractable.
        self.mht.boards = self.mht.boards[:3_000]
        # The true board could be the board that results from any possible move on each board that
        # was tracked before the move. That results in a growth factor of roughly 30, and higher
        # still in the late game when the board is more open.
        self.mht.op_move(capture_square)

    def choose_sense(
        self,
        sense_actions: List[chess.Square],
        move_actions: List[chess.Move],
        seconds_left: float,
    ) -> Optional[chess.Square]:
        if not self.mht.boards:
            return None
        # You can think of the choice of a sense square as a partition of the possible boards,
        # where possible boards are equivalent if the sense result centered on that square is the
        # same. As such, it is possible to identify choices that are dominated (they can't
        # possibly yield information that would not be revealed by the dominating choice) in
        # terms of these partitions. Additionally, it is easy to write simple logic to recommend
        # a sense choice based on the partitions. For example, the following function recommends
        # the square whose biggest partition is smallest (the minimax remaining number of boards
        # after the hypothetical sense step).
        self.mht.speculate_sense(non_dominated_sense_by_own_pieces(self.mht.boards[0]))
        minimax_square = minimax_sense(self.mht.sense_speculation)
        return minimax_square

    def handle_sense_result(
        self, sense_result: List[Tuple[chess.Square, Optional[chess.Piece]]]
    ):
        if sense_result:  # False if we skipped sensing
            # The sense target is just the middle of the list when in sorted order. Note that the
            # sense result from our simulate_sense utility and from the reconchess LocalGame are
            # already in sorted order, but that sense results from server games will not necessarily
            # be sorted. Also, the square, piece pairs are tuples except in server games, in which
            # case they are lists and we convert them to tuples here.
            sorted_result = sorted((s, p) for s, p in sense_result)
            square = sorted_result[4][0]
            # Here we filter the list of possible boards to include only those that match over the
            # sensed region.
            self.mht.sense(square, sorted_result)

    def choose_move(
        self, move_actions: List[chess.Move], seconds_left: float
    ) -> Optional[chess.Move]:
        # Since we limit the size of the MHT board list, it is possible for that list to become
        # empty. In that case we fall back to requesting moves randomly.
        if not self.mht.boards:
            return random.choice(move_actions)
        # If any move is guaranteed to result in a king capture, take it!
        winning_move = certain_win(self.mht.boards)
        if winning_move:
            return winning_move
        # Otherwise, let stockfish evaluate over all possible boards and tally a vote.
        move = vote(move_actions, self.mht.boards, self.engine)
        # The reconchess library encodes passing (null) moves as None so we convert
        # chess.Move.null() to None here using the fact that it evaluates to truthy false.
        return move or None

    def handle_move_result(
        self,
        requested_move: Optional[chess.Move],
        taken_move: Optional[chess.Move],
        captured_opponent_piece: bool,
        capture_square: Optional[chess.Square],
    ):
        # Here we update all possible boards by pushing our move to them and only keeping the ones
        # with the same capture square as observed.
        self.mht.move(
            requested_move or chess.Move.null(),
            taken_move or chess.Move.null(),
            capture_square,
        )

    def handle_game_end(
        self,
        winner_color: Optional[Color],
        win_reason: Optional[WinReason],
        game_history: GameHistory,
    ):
        self.engine.close()


def vote(possible_requested_moves, boards, engine):
    # This is just one of the many ways to aggregate the perfect-information recommendations over
    # each possible board into a move decision. Additionally, the general approach of aggregating
    # recommendations over MHT hypotheses is not necessarily the best strategy.
    # Imperfect-information-native approaches that don't involve MHT (like
    # counterfactual-regret-minimization) should theoretically outperform MHT approaches, though
    # through the first two competitions MHT bots such as StrangeFish and Oracle continue to be top
    # performers. (Also, MHT can be helpful in CFR and similar approaches, for example by
    # identifying dominated actions.)
    #
    # In this function, we choose a move to request by letting stockfish place votes over a random
    # subset of the MHT boards and selecting a winner by ranked-choice-voting. We must separately
    # handle unusual board configurations (i.e. the opponent king can be captured, or we are in
    # checkmate when it is our turn to move) and we let stockfish rank move options on the rest.
    # Because sometimes multiple move requests would be amended to the same taken move, we allow
    # choices to have a tied rank and nominate all such move requests with the same rank as the
    # taken move suggested by stockfish. In the unusual case where we have multiple options for
    # capturing the opponent king, this also gives us a way to nominate all those options as equal
    # first choices.
    votes = []
    random.shuffle(boards)
    for board in tqdm(boards[:1200]):
        my_ranked_votes = []
        votes.append(my_ranked_votes)
        # All requested moves that result in the voted-for taken moves are counted equally.
        move_lookup = defaultdict(list)
        for requested_move in possible_requested_moves:
            taken_move, _ = simulate_move(board, requested_move)
            move_lookup[taken_move].append(requested_move)
        # Boards where the king can be captured cannot be scored by stockfish.
        # Instead, vote equally for all possible king capture moves.
        op_king_square = board.king(not board.turn)
        king_attackers = board.attackers(board.turn, op_king_square)
        if king_attackers:
            my_ranked_votes.append([])
            for attacker in king_attackers:
                taken_move = chess.Move(attacker, op_king_square)
                requested_moves = move_lookup[taken_move]
                my_ranked_votes[0] += requested_moves
        else:
            board.clear_stack()
            results = engine.analyse(
                board, limit=chess.engine.Limit(depth=8), multipv=4
            )
            for result in results:
                try:
                    taken_move = result["pv"][0]
                    my_ranked_votes.append(move_lookup[taken_move])
                except KeyError:
                    pass  # No moves were suggested because we are in checkmate on this board.
    # Ranked-choice-voting is an iterative algorithm that scores candidates by the number of
    # first-choice votes they receive. If a candidate receives a majority, it is selected.
    # Otherwise, the lowest-scoring candidate is eliminated and the process repeats. Because this
    # version allows tied ranking, the total number of votes can exceed the number of voters.
    while True:
        if not votes:
            return chess.Move.null()
        threshold = len(votes) // 2
        first_choice_votes = defaultdict(int)
        for vote in votes:
            for move in vote[0]:
                first_choice_votes[move] += 1
        max_move, max_num_votes = max(first_choice_votes.items(), key=lambda x: x[1])
        if max_num_votes >= threshold:
            return max_move
        min_move, min_num_votes = min(first_choice_votes.items(), key=lambda x: x[1])
        revised_votes = []
        for vote in votes:
            revised_vote = []
            for group in vote:
                revised_group = [move for move in group if move != min_move]
                if revised_group:
                    revised_vote.append(revised_group)
            if revised_vote:
                revised_votes.append(revised_vote)
        votes = revised_votes



from typing import Dict, List, Optional, Tuple

import chess
from reconchess.utilities import move_actions, revise_move

from reconchess_tools.utilities import simulate_move

# Sensing on the edge of the board is never a good idea
SENSE_SQUARES = [
    square
    for square in chess.SQUARES
    if 0 < chess.square_file(square) < 7 and 0 < chess.square_rank(square) < 7
]


def non_dominated_moves(boards: List[chess.Board]):
    move_choices = {chess.Move.null()}
    for requested_move in move_actions(boards[0]):
        for board in boards:
            if requested_move == revise_move(board, requested_move):
                move_choices.add(requested_move)
                break
    return move_choices


def certain_win(boards: List[chess.Board]) -> Optional[chess.Move]:
    for requested_move in move_actions(boards[0]):
        for board in boards:
            op_king_square = board.king(not board.turn)
            if requested_move.to_square != op_king_square:
                break  # this can't possibly be a king-capture
            _, capture_square = simulate_move(board, requested_move)
            if capture_square != op_king_square:
                break  # this isn't a king-capture
        else:
            return requested_move


def minimax_sense(
    sense_results_for_square: Dict[chess.Square, Dict[Tuple, chess.Board]]
):
    """Find the minimax sense square

    Returns the square for which the worst case number of boards remaining after sensing there is
    the smallest.
    """
    return min(
        sense_results_for_square.items(),
        key=lambda x: max(len(group) for group in x[1].values()),
    )[0]


def non_dominated_sense(
    sense_results_for_square: Dict[chess.Square, Dict[Tuple, chess.Board]]
):
    # a square is dominated if every group in its groups is a superset of some group of the dominating square
    dominated_senses = set()
    for square, sense_results in sense_results_for_square.items():
        # Assume equal boards are identical objects as they are with mht.speculate_sense. This leads
        # to drastically faster comparisons than if we had to check the boards for equality.
        groups = [set(id(board) for board in v) for v in sense_results.values()]
        for square2, sense_results2 in sense_results_for_square.items():
            if square2 in dominated_senses or square2 == square:
                continue
            groups2 = [set(id(board) for board in v) for v in sense_results2.values()]
            if all(any(g.issuperset(g2) for g2 in groups2) for g in groups):
                dominated_senses.add(square)
                break
    return set(sense_results_for_square.keys()) - dominated_senses


def non_dominated_sense_by_own_pieces(board):
    """Screen sense squares that are trivially dominated because of my own piece positions"""
    actions = []
    for square in SENSE_SQUARES:
        rank = chess.square_rank(square)
        file = chess.square_file(square)
        # Skip if can sense higher and bottom row is full of my pieces
        if rank < 6:
            for offset in [-9, -8, -7]:
                piece = board.piece_at(square + offset)
                if piece is None or piece.color != board.turn:
                    break
            else:
                continue
        # Skip if can sense lower and top row is full of my pieces
        if rank > 1 and square - 8 in actions:
            for offset in [7, 8, 9]:
                piece = board.piece_at(square + offset)
                if piece is None or piece.color != board.turn:
                    break
            else:
                continue
        # Skip if can sense right and left col is full of my pieces
        if file < 6:
            for offset in [-9, -1, 7]:
                piece = board.piece_at(square + offset)
                if piece is None or piece.color != board.turn:
                    break
            else:
                continue
        # Skip if can sense left and right col is full of my pieces
        if file > 1 and square - 1 in actions:
            for offset in [-7, 1, 9]:
                piece = board.piece_at(square + offset)
                if piece is None or piece.color != board.turn:
                    break
            else:
                continue
        actions.append(square)
    return actions
